#include "SendRecvThread.h"
#include "MySyncObject.h"		// CMySyncObjectを使うため
#include "RingBuff.h"			// CRingBuffを使うため
#include "RecvFileThread.h"		// CRecvFileThreadを使うため
#include "SendFileThread.h"		// ★CSendFileThreadを使うため
#include "ChatMsgThread.h"		// CChatMsgThreadを使うため

// 送信速度
#define		SEND_BPS	(1000000.0)	// 送信速度1Mbps

//----------------------------------------------
// function
//    コンストラクタ
// parameter
//    ConnectionInfoRec *pConInfo	[in]機能に必要な情報
// return
//    なし
//----------------------------------------------
CSendRecvThread::CSendRecvThread(ConnectionInfoRec *pConInfo)
{
	m_pConInfo = pConInfo;
	m_fIamZombie = FALSE;
	m_pCRingBuffSend = new CRingBuff(SEND_BUFF_SIZE);	// 送信リングバッファの構築
	m_pCRingBuffRecv = new CRingBuff(RECV_BUFF_SIZE);	// 受信リングバッファの構築

	m_dwPrevSentTime = 0;								// 初回送信はすぐに
	m_dwSendInterval = 0;								// 

	m_pCRecvFileThread = new CRecvFileThread(this);		// ファイル受信スレッドを常に使えるように準備する
	m_pCRecvFileThread->Begin();

	m_pCSendFileThread = new CSendFileThread(this);		// ★ファイル送信スレッドを常に使えるように準備する
	m_pCSendFileThread->Begin();

	m_pCChatMsgThread = new CChatMsgThread(this);		// チャットメッセージ処理を常に使えるように準備する
	m_pCChatMsgThread->Begin();
}

//----------------------------------------------
// function
//    デストラクタ
// parameter
//    なし
// return
//    なし
//----------------------------------------------
CSendRecvThread::~CSendRecvThread()
{
	if (m_pCChatMsgThread != NULL)						// このクラスが破棄されるときにチャットメッセージ処理スレッドを破棄
	{
		m_pCChatMsgThread->End();
		m_pCChatMsgThread->WaitForEnd();
	}
	SAFE_DELETE(m_pCChatMsgThread)
	if (m_pCRecvFileThread != NULL)						// このクラスが破棄されるときにファイル受信スレッドを破棄
	{
		m_pCRecvFileThread->End();
		m_pCRecvFileThread->WaitForEnd();
	}
	SAFE_DELETE(m_pCRecvFileThread)
	if (m_pCSendFileThread != NULL)						// ★このクラスが破棄されるときにファイル送信スレッドを破棄
	{
		m_pCSendFileThread->End();
		m_pCSendFileThread->WaitForEnd();
	}
	SAFE_DELETE(m_pCRecvFileThread)

	SAFE_DELETE(m_pCRingBuffSend)		// 送信リングバッファの破棄
	SAFE_DELETE(m_pCRingBuffRecv)		// 受信リングバッファの破棄
}

//----------------------------------------------
// function
//    機能を記述した関数
// parameter
//    なし
// return
//    0:正常 -1:エラー発生
//----------------------------------------------
UINT CSendRecvThread::DoWork()
{
	BOOL		fRet = TRUE;
	char	    szRecvBuffer[RCVBUFSIZE];	// 受信バッファ
	int			iRecvSize;
	pollfd		fds[1] = { 0 };
	char		*pcData = NULL;				// 未送信データ
	int			iSendSize = 0;				// 未送信データサイズ
	DWORD		dwNow;						// 送信チェックした時刻を覚えるため

	fprintf(stderr, "DoWork()\n");
	fds[0].fd = m_pConInfo->fdClient;
	while (!m_fStopFlag)
	{
		// ★致命的エラー発生のスレッドを破棄したときは終了
		// 終了すると当該クライアントは切断される
		if (KillZombei() == TRUE)
		{
			fRet = FALSE;
			break;
		}

		// 送信データがあるかリングバッファを調べる
		iSendSize = m_pCRingBuffSend->GetReadableSize();
		fds[0].events = POLLIN | POLLRDHUP;					// 受信と相手側からの切断イベントを設定
		// チェックのために現在時刻を取得する
		dwNow = timeGetTime();
		// 送信可能時刻かつ未送信データがあるときだけ送信可能検査
		if ((iSendSize > 0) && (CanSendNow(dwNow) == TRUE))
			fds[0].events |= POLLOUT;						// 書き込み可能を追加

		poll(fds, 1, 10);
		if (fds[0].revents & POLLRDHUP)			// 相手側からの切断
		{
			fprintf(stderr, "Disconnected pollrdhup\n");
			fRet = FALSE;
			break;
		}
		if (fds[0].revents & POLLERR)			// エラー発生
		{
			m_pConInfo->pCMySyncObject->Lock();
			DispErrorMsg("Err:DoWork");
			m_pConInfo->pCMySyncObject->UnLock();
			fRet = FALSE;
			break;
		}
		// 受信処理
		if (fds[0].revents & POLLIN)			// 受信データイベント
		{
			// 受信リングバッファに空きがあれば受信する
			// ここで受信しなければ、次回のrevents検査で受信が行われる
			iRecvSize = min(m_pCRingBuffRecv->GetWriteableSize(), RCVBUFSIZE);
			if (iRecvSize > 0)
			{
				if ((iRecvSize = recv(m_pConInfo->fdClient, szRecvBuffer, iRecvSize, 0)) <= 0)
				{
					m_pConInfo->pCMySyncObject->Lock();
					if (iRecvSize == 0)
						DispErrorMsg("Disconnected recv");
					else
						DispErrorMsg("Err:recv");
					m_pConInfo->pCMySyncObject->UnLock();
					fRet = FALSE;
					break;
				}
				else
				{
					// 取得したデータすぐに受信リングバッファに書きこむ
					// 受信リングバッファに書き込むのはこのスレッドだけなので、すべて書き込めるはず
					m_pCRingBuffRecv->Write((LPBYTE)szRecvBuffer, iRecvSize);
				}
			}
		}
		// 受信リングバッファに格納されているデータの解析を行う
		// 複数のパケットが格納されている可能性があるので、reventsの結果とは無関係に
		// 解析を行うようにする
		if (AnalyzeDataRecv() == -1)
		{
			m_pConInfo->pCMySyncObject->Lock();
			DispErrorMsg("Err:Packet format");
			m_pConInfo->pCMySyncObject->UnLock();
			fRet = FALSE;
			break;
		}
		// 送信処理
		if (fds[0].revents & POLLOUT)		// 送信可能ならsend実施
		{
			// 送信したいデータの取得(PATH_MTUより小さくなるように取得する)
			iSendSize = GetSendData(&pcData);
			if (send(m_pConInfo->fdClient, pcData, iSendSize, 0) != iSendSize)
			{
				DispErrorMsg("Err:send");
				fRet = FALSE;
				break;
			}
			SAFE_FREE(pcData)
			// 送信時刻と次回送信までの時間をセット
			m_dwPrevSentTime = dwNow;
			m_dwSendInterval = CalcNextSendInterval(iSendSize);
		}
	}
	m_pConInfo->pCMySyncObject->Lock();
	m_fIamZombie = TRUE;
	m_pConInfo->pCMySyncObject->UnLock();
	SAFE_FREE(pcData)
	return((fRet == TRUE) ? 0 : -1);
}

//----------------------------------------------
// function
//		送信データの設定
//		送信できない時、エラーとして切断をするのが良いか
//		は迷うところ
// parameter
//		char	*pcData	[in]送信データ
//		int		iSize	[in]データ長
// return
//		TRUE/FALSE
//----------------------------------------------
BOOL CSendRecvThread::SetSendData(char *pcData, int iSize)
{
	BOOL	fRet = FALSE;

	// 送信リングバッファに空きがないときは書き込まない
	fRet = m_pCRingBuffSend->Write((LPBYTE)pcData, iSize);
	return(fRet);
}

//----------------------------------------------
// function
//		送信データの取得
// parameter
//		char **ppcData	[in/out]送信データ
// return
//		データ長
//----------------------------------------------
int CSendRecvThread::GetSendData(char **ppcData)
{
	int		iSize = 0;

	// 送信データがあるかリングバッファのデータサイズを調べる
	if ((iSize = m_pCRingBuffSend->GetReadableSize()) > 0)
	{
		// 送信サイズをPATH_MTUより小さくしておく(1024+共通ヘッダなら大丈夫)
		iSize = min(iSize, SENDBUFSIZE);
		*ppcData = (char *)calloc(iSize, sizeof(char));
		iSize = m_pCRingBuffSend->Read((LPBYTE)*ppcData, iSize);
	}
	return(iSize);
}

//----------------------------------------------
// function
//		チャット分配用データの取得
//		パケット単位で取り出す
//		今回は１パケット分
// parameter
//		char **ppcData	[in/out]チャット転送用データ
// return
//		データ長
//----------------------------------------------
int CSendRecvThread::GetDataForDistributeChat(char **ppcData)
{
	int			iSize = 0;

	if (m_pCChatMsgThread != NULL)
		iSize = m_pCChatMsgThread->GetDataForDistributeChat(ppcData);
	return(iSize);
}


//----------------------------------------------
// function
//    このスレッドはゾンビ状態か
//	  別スレッドから参照される
// parameter
//    なし
// return
//    0:正常 -1:エラー発生
//----------------------------------------------
BOOL CSendRecvThread::IsZombie()
{
	BOOL	fRet;

	m_pConInfo->pCMySyncObject->Lock();
	fRet = m_fIamZombie;
	m_pConInfo->pCMySyncObject->UnLock();
	return(fRet);
}

//----------------------------------------------
// function
//		受信データの解析
//		受信リングバッファに格納されているデータを調べる
// parameter
//		なし
// return
//		0:パケットが揃っていない
//		1:パケットを受信したので処理を行った
//		-1:エラーが発生した
//----------------------------------------------
BOOL CSendRecvThread::AnalyzeDataRecv()
{
	int			iRet = 0;
	HeaderRec	Header;
	WORD		wCmd;
	// データサイズを調べる
	int			iSize = m_pCRingBuffRecv->GetReadableSize();

	if (iSize < sizeof(HeaderRec))			// ヘッダサイズに満たないときは何もしない
		goto L_END;

	// ヘッダ部を借り読み込みする
	m_pCRingBuffRecv->ReadWithoutUpdateHeadPoint((LPBYTE)&Header, sizeof(HeaderRec));
	// ヘッダ部の解析
	if (memcmp(Header.bMagicData, MAGIC_STRING, strlen(MAGIC_STRING)) != 0)
	{
		iRet = -1;							// 識別子が違うのでエラー
		goto L_END;
	}
	wCmd = ntohs(Header.wCommand);
	Locate(1, 10, 1);
	fprintf(stderr, "CMD:%d", wCmd);

	switch (wCmd)
	{
		case CMD_MSG_DATA:
			iRet = RecvMessagePacket(&Header);
			break;
		// ファイル送信コマンド(C->S)を処理する
		case CMD_SEND_FILE_START_REQ:
		case CMD_SEND_FILE_ING:
		case CMD_SEND_FILE_END_REQ:
		case CMD_SEND_FILE_ABORT_REQ:
			iRet = RecvFileSendPacket(&Header);
			break;
		// ★ファイル受信コマンド(C->S)を処理する
		case CMD_RECV_FILE_SIZE_REQ:
		case CMD_RECV_FILE_START_REQ:
		case CMD_RECV_FILE_END_RES:
		case CMD_RECV_FILE_ABORT_REQ:
			iRet = RecvFileRecvPacket(&Header);
			break;
		default:						// 知らないコマンド
			iRet = -1;
			break;
	}
L_END:
	return(iRet);
}

//----------------------------------------------
// function
//		メッセージコマンドの受信
// parameter
//		HeaderRec *pHeader	[in]仮読みしたヘッダ
// return
//		0:パケットが揃っていない
//		1:パケットを受信したので処理を行った
//		-1:エラーが発生した
//----------------------------------------------
BOOL CSendRecvThread::RecvMessagePacket(HeaderRec *pHeader)
{
	int			iRet = 0;
	int			iDataSize, iSize, iPacketSize;
	LPBYTE		pbPacket = NULL;

	iDataSize = ntohs(pHeader->wDataLen);
	iPacketSize = iDataSize + sizeof(HeaderRec);

	iSize = m_pCRingBuffRecv->GetReadableSize();
	if (iSize < iPacketSize)
		goto L_END;
	pbPacket = (LPBYTE)calloc(iPacketSize, sizeof(BYTE));
	// パケット全体を仮読み
	m_pCRingBuffRecv->ReadWithoutUpdateHeadPoint(pbPacket, iPacketSize);
	if (m_pCChatMsgThread != NULL)
	{
		// チャットメッセージ処理スレッドに渡せたら読み込んだことにする
		if (m_pCChatMsgThread->m_pCRingBuffCmd->Write(pbPacket, iPacketSize) == TRUE)
		{
			m_pCRingBuffRecv->UpdateHeadPoint(iPacketSize);
			iRet = 1;
		}
	}
	else
	{
		// チャットメッセージ処理スレッドがいなければエラー切断
		m_pCRingBuffRecv->UpdateHeadPoint(iPacketSize);
		iRet = -1;
	}
	SAFE_FREE(pbPacket);
L_END:
	return(iRet);
}

//----------------------------------------------
// function
//		ファイル送信コマンドの受信
//		ファイル送信コマンド(C->S)をCRecvThreadに渡すだけ
// parameter
//		HeaderRec *pHeader	[in]仮読みしたヘッダ
// return
//		0:パケットが揃っていない, ファイル受信側が満杯
//		1:パケットを受信したので処理を行った
//		-1:エラーが発生した
//----------------------------------------------
int CSendRecvThread::RecvFileSendPacket(HeaderRec *pHeader)
{
	int			iRet = 0;
	int			iDataSize, iSize, iPacketSize;
	LPBYTE		pbPacket = NULL;

	iDataSize = ntohs(pHeader->wDataLen);
	iPacketSize = iDataSize + sizeof(HeaderRec);

	iSize = m_pCRingBuffRecv->GetReadableSize();
	if (iSize < iPacketSize)
		goto L_END;
	pbPacket = (LPBYTE)calloc(iPacketSize, sizeof(BYTE));
	// パケット全体を仮読み
	m_pCRingBuffRecv->ReadWithoutUpdateHeadPoint(pbPacket, iPacketSize);
	if (m_pCRecvFileThread != NULL)
	{
		// ファイル受信スレッドに渡せたら読み込んだことにする
		if (m_pCRecvFileThread->m_pCRingBuffCmd->Write(pbPacket, iPacketSize) == TRUE)
		{
			m_pCRingBuffRecv->UpdateHeadPoint(iPacketSize);
			iRet = 1;
		}
	}
	else
	{
		// ファイル受信スレッドがいなければエラー切断
		m_pCRingBuffRecv->UpdateHeadPoint(iPacketSize);
		iRet = -1;
	}
	SAFE_FREE(pbPacket);
L_END:
	return(iRet);
}

//----------------------------------------------
// function
//		★ファイル受信コマンドの受信
//		ファイル受信コマンド(C->S)をCSendThreadに渡すだけ
// parameter
//		HeaderRec *pHeader	[in]仮読みしたヘッダ
// return
//		0:パケットが揃っていない, ファイル受信側が満杯
//		1:パケットを受信したので処理を行った
//		-1:エラーが発生した
//----------------------------------------------
int CSendRecvThread::RecvFileRecvPacket(HeaderRec *pHeader)
{
	int			iRet = 0;
	int			iDataSize, iSize, iPacketSize;
	LPBYTE		pbPacket = NULL;

	iDataSize = ntohs(pHeader->wDataLen);
	iPacketSize = iDataSize + sizeof(HeaderRec);

	iSize = m_pCRingBuffRecv->GetReadableSize();
	if (iSize < iPacketSize)
		goto L_END;
	pbPacket = (LPBYTE)calloc(iPacketSize, sizeof(BYTE));
	// パケット全体を仮読み
	m_pCRingBuffRecv->ReadWithoutUpdateHeadPoint(pbPacket, iPacketSize);
	if (m_pCSendFileThread != NULL)
	{
		// ファイル送信スレッドに渡せたら読み込んだことにする
		if (m_pCSendFileThread->m_pCRingBuffCmd->Write(pbPacket, iPacketSize) == TRUE)
		{
			m_pCRingBuffRecv->UpdateHeadPoint(iPacketSize);
			iRet = 1;
		}
	}
	else
	{
		// ファイル受信スレッドがいなければエラー切断
		m_pCRingBuffRecv->UpdateHeadPoint(iPacketSize);
		iRet = -1;
	}
	SAFE_FREE(pbPacket);
L_END:
	return(iRet);
}


//----------------------------------------------
// function
//		次の送信までの間隔(msec)を求める
// parameter
//		int iSentSize [in]送信したサイズ
// return
//		msec
//----------------------------------------------
DWORD CSendRecvThread::CalcNextSendInterval(int iSentSize)
{
	DWORD	dwInterval = 0;
	double	dbBytePerSec = SEND_BPS / 8.0;
	double	dbCountPerSec;

	if (iSentSize == 0)
		goto L_END;
	dbCountPerSec = dbBytePerSec / (double)iSentSize;
	dwInterval = DWORD(1000.0 / dbCountPerSec);

L_END:
	return(dwInterval);
}


//----------------------------------------------
// function
//		今送信可能時刻か
// parameter
//		DWORD dwNow		[in]現在時刻
// return
//		TRUE/FALSE
//----------------------------------------------
BOOL CSendRecvThread::CanSendNow(DWORD dwNow)
{
	BOOL	fRet = FALSE;

	if (GetdwInterval(dwNow, m_dwPrevSentTime) >= m_dwSendInterval)
		fRet = TRUE;

	return(fRet);
}

//----------------------------------------------
// function
//		★このスレッドが作ったスレッドが無効(致命的エラー)の時破棄
//		破棄したときは切断処理のために CSendRecvThreadも終わる
// parameter
//		なし
// return
//		TRUE:致命的エラーで破棄した/FALSE:致命的エラーなし
//----------------------------------------------
BOOL CSendRecvThread::KillZombei()
{
	BOOL	fRet = FALSE;

	if (m_pCRecvFileThread != NULL)
	{
		if (m_pCRecvFileThread->IsZombie() == TRUE)
		{
			m_pCRecvFileThread->End();
			m_pCRecvFileThread->WaitForEnd();
			SAFE_DELETE(m_pCRecvFileThread)
			fRet = TRUE;
		}
	}
	if (m_pCSendFileThread != NULL)			// ★
	{
		if (m_pCSendFileThread->IsZombie() == TRUE)
		{
			m_pCSendFileThread->End();
			m_pCSendFileThread->WaitForEnd();
			SAFE_DELETE(m_pCSendFileThread)
			fRet = TRUE;
		}
	}
	if(m_pCChatMsgThread != NULL)
	{
		if (m_pCChatMsgThread->IsZombie() == TRUE)
		{
			m_pCChatMsgThread->End();
			m_pCChatMsgThread->WaitForEnd();
			SAFE_DELETE(m_pCChatMsgThread)
			fRet = TRUE;
		}
	}
	return(fRet);
}
